@@ -6,7 +6,7 @@ exports.chainableExec = chainableExec
 exports.whichAndExec = whichAndExec
 
 var exec = require("child_process").execFile
-  , spawn = require("child_process").spawn
+  , spawn = require("./spawn")
   , npm = require("../npm.js")
   , which = require("which")
   , git = npm.config.get("git")
